from .base_api import base_api

class twitter_recsys(base_api):
    def __init__(self):
        super().__init__()
        self.saved_path = self.download_s3("pyrecdp-testdata", "test_twitter_recsys.parquet")

    def to_pandas(self, nrows = None):
        import pandas as pd
        return pd.read_parquet(self.saved_path)
         


from pyrecdp.primitives.generators import *
from pyrecdp.core import DataFrameSchema, SparkDataProcessor, DiGraph
from pyrecdp.primitives.operations import Operation, DataFrameOperation, RDDToDataFrameConverter
import pandas as pd
import logging
import graphviz
import json
from pyrecdp.core.utils import Timer, sample_read
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD as SparkRDD

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class BasePipeline:
    def __init__(self, dataset, label, *args, **kwargs):
        if isinstance(dataset, pd.DataFrame):
            self.dataset = {'main_table': dataset}
        elif isinstance(dataset, list):
            self.dataset = dict((idx, data) for idx, data in enumerate(dataset))
        elif isinstance(dataset, dict):
            self.dataset = dataset
        else:
            self.dataset = {'main_table': dataset}
        main_table = None
        input_is_path = False
        if isinstance(label, str):    
            for data_key, data in self.dataset.items():
                # optional: data is file_path
                if isinstance(data, str):
                    input_is_path = True
                    data = sample_read(data)
                if label in data.columns:
                    main_table = data_key
                    break
        if not main_table:
            raise ValueError(f"label {label} is not found in dataset")
        
        # Set properties for BasePipeline
        if not input_is_path:
            original_data = self.dataset[main_table]
        else:
            original_data = sample_read(self.dataset[main_table])
        y = original_data[label]
        to_select = [i for i in original_data.columns if i != y.name]
        self.feature_data = original_data[to_select]
            
        self.generators = []
        self.pipeline = DiGraph()
        if not input_is_path:
            self.pipeline[0] = Operation(
                0, None, output = DataFrameSchema(self.feature_data), op = 'DataFrame', config = main_table)
        else:
            self.pipeline[0] = Operation(
                0, None, output = DataFrameSchema(self.feature_data), op = 'DataLoader', config = {'table_name': main_table, 'file_path': self.dataset[main_table]})

        if len(self.dataset) > 1:
            self.supplementary = dict((k, v) for k, v in self.dataset.items() if k != main_table)
        else:
            self.supplementary = None
        self.rdp = None
    
    def fit_analyze(self, *args, **kwargs):
        child = list(self.pipeline.keys())[-1]
        max_id = child
        for i in range(len(self.generators)):
            for generator in self.generators[i]:
                self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id)

    def __repr__(self):
        return repr(self.pipeline)

    def export(self, file_path = None):
        json_object = self.pipeline.json_dump()
        if file_path:
            # Writing to sample.json
            with open(file_path, "w") as outfile:
                outfile.write(json_object)
        else:
            print(json_object)
    
    def import_from_json(self, file_path):
        with open(file_path, "r") as infile:
            json_object = json.load(infile)
        for idx, op_config in json_object.items():
            idx = int(idx)
            if idx in self.pipeline:
                continue
            self.pipeline[idx] = Operation.load(idx, op_config)
        #self.create_executable_pipeline()
        
    def create_executable_pipeline(self):
        node_chain = self.pipeline.convert_to_node_chain()
        executable_pipeline = DiGraph()
        executable_sequence = []
        for idx in node_chain:
            executable_pipeline[idx] = self.pipeline[idx].instantiate()
            executable_sequence.append(executable_pipeline[idx])
        return executable_pipeline, executable_sequence
        
    def plot(self):
        f = graphviz.Digraph()
        edges = []
        nodes = []
        f.attr(fontsize='10')
        def add_escape(input):
            input = input.replace('<', '\<').replace('>', '\>')
            #input = input.replace("'", "\\\'").replace("\"", "\\\"")
            return input

        def add_break(input):
            if isinstance(input, list) and len(input) < 3:
                for line in input:
                    if isinstance(line, str):
                        ret = str(input)
                        return ret
            if isinstance(input, dict):
                input = [f"{k}: {add_break(v)}" for k, v in input.items()]
            if isinstance(input, list):
                ret = ""
                for line in input:
                    ret += str(add_break(line)) + "\l"
                return ret
            return input

        for node_id, config in self.pipeline.items():
            nodes.append([str(node_id), f"{config.op} |{add_escape(str(add_break(config.config)))}"])
            if config.children:
                for src_id in config.children:
                    edges.append([str(src_id), str(node_id)])
        for node in nodes:
            f.node(node[0], node[1], shape='record', fontsize='12')
        for edge in edges:
            f.edge(*edge)
        return f  

    def to_chain(self):
        return self.pipeline.convert_to_node_chain()
        
    def execute(self, engine_type = "pandas"):
        # prepare pipeline
        executable_pipeline, executable_sequence = self.create_executable_pipeline()

        # execute
        if engine_type == 'pandas':
            with Timer(f"execute with pandas"):
                for op in executable_sequence:
                    if isinstance(op, DataFrameOperation):
                        op.set(self.dataset)
                    with Timer(f"execute {op}"):
                        op.execute_pd(executable_pipeline)
            df = executable_sequence[-1].cache
        elif engine_type == 'spark':
            for op in executable_sequence:
                if isinstance(op, DataFrameOperation):
                    op.set(self.dataset)
                print(f"append {op}")
                op.execute_spark(executable_pipeline, self.rdp)
            ret = executable_sequence[-1].cache
            if isinstance(ret, SparkDataFrame):
                with Timer(f"execute with spark"):
                    df = self.rdp.transform(ret)
            elif isinstance(ret, SparkRDD):
                _convert = RDDToDataFrameConverter().get_function(self.rdp)
                with Timer(f"execute with spark"):
                    df = _convert(ret)
            else:
                raise ValueError(f"unrecognized {ret} produced by execute with spark")
        else:
            raise NotImplementedError('pipeline only support pandas and spark as engine')
        
        # fetch result
        return df

    def fit_transform(self, engine_type = 'pandas', *args, **kwargs):
        if engine_type == "spark":
            self.rdp = SparkDataProcessor()
        ret = self.execute(engine_type)
        if engine_type == "spark":
            del self.rdp 
        return ret


from .base import BaseOperation
import pandas as pd
from pyspark.sql import DataFrame as SparkDataFrame
import copy

class CategorifyOperation(BaseOperation):
    def __init__(self, op_base):
        super().__init__(op_base)
        self.feature_in = op_base.config
        self.support_spark_dataframe = False
        self.support_spark_rdd = True
    
    def get_function_pd(self):
        feature_in = copy.deepcopy(self.feature_in)
        def categorify(df):
            for feature in feature_in:
                codes, uniques = pd.factorize(df[feature])
                df[f"{feature}__idx"] = pd.Series(codes, df[feature].index)
            return df
        return categorify


from .base_api import base_api

class nyc_taxi(base_api):
    def __init__(self, scale = 'test'):
        super().__init__()
        self.scale = scale
        if scale == 'test':            
            self.saved_path = self.download_s3("pyrecdp-testdata", "test_nyc_taxi_fare.parquet")
        elif scale == 'test_large':
            self.saved_path = self.download_s3("pyrecdp-testdata", "nyc_taxi_fare_1M.csv")
        elif scale == 'full':
            self.name = "nyc_taxi_fare_cleaned.csv"
            self.url = "https://huggingface.co/datasets/Chendi/NYC_TAXI_FARE_CLEANED/resolve/main/nyc_taxi_fare_cleaned.csv"
            
            self.saved_path = self.download_url(self.name, self.url)

    def to_pandas(self, nrows = None):
        import pandas as pd
        if self.scale == 'test':            
            return pd.read_parquet(self.saved_path)
        elif self.scale == 'test_large':
            return pd.read_csv(self.saved_path)
        elif self.scale == 'full':            
            if nrows:
                return pd.read_csv(self.saved_path, nrows = nrows)
            else:
                return pd.read_csv(self.saved_path)
         


import json
import pandas as pd
from .dataframe import *
from pyspark.sql import DataFrame as SparkDataFrame
from pyspark import RDD as SparkRDD
from pyrecdp.core.utils import dump_fix

class Operation:
    def __init__(self, idx, children, output, op, config):
        self.idx = idx
        self.children = children #input operation
        self.output = output #output schema
        self.op = op #func name
        self.config = config #operation config

    def __repr__(self):
        return repr(self.dump())
    
    def dump(self):
        dump_dict = {
            #'idx': self.idx,
            'children': self.children,
            #'output': self.output,
            'op': self.op,
            'config': dump_fix(self.config)
        }
        return dump_dict
    
    def instantiate(self):
        from .data import DataFrameOperation, DataLoader
        from .merge import MergeOperation
        from .name import RenameOperation
        from .category import CategorifyOperation
        from .datetime import DatetimeOperation
        from .drop import DropOperation
        from .fillna import FillNaOperation
        from .geograph import HaversineOperation
        from .nlp import BertDecodeOperation, TextFeatureGenerator
        from .type import TypeInferOperation
        from .tuple import TupleOperation

        operations_ = {
            'DataFrame': DataFrameOperation,
            'DataLoader': DataLoader,
            'merge': MergeOperation,
            'rename': RenameOperation,
            'categorify': CategorifyOperation,
            'datetime_feature': DatetimeOperation,
            'drop': DropOperation,
            'fillna': FillNaOperation,
            'haversine': HaversineOperation,
            'tuple': TupleOperation,
            'bert_decode': BertDecodeOperation,
            'text_feature': TextFeatureGenerator,
            'type_infer': TypeInferOperation
        }

        if self.op in operations_:
            return operations_[self.op](self)
        else:
            raise NotImplementedError(f"operation {self.op} is not implemented.")
 
    @staticmethod
    def load(idx, dump_dict):
        obj = Operation(idx, dump_dict['children'], None, dump_dict['op'], dump_dict['config'])
        return obj

class BaseOperation:
    def __init__(self, op_base):
        self.op = op_base
        self.cache = None
        self.support_spark_dataframe = False
        self.support_spark_rdd = False
       
    def __repr__(self) -> str:
        return self.op.op
        
    def execute_pd(self, pipeline):
        if self.cache is not None:
            return
        _proc = self.get_function_pd()
        if not self.op.children or len(self.op.children) == 0:
            self.cache = _proc()
        else:
            child_output = pipeline[self.op.children[0]].cache
            self.cache = _proc(child_output)
            
    def execute_spark(self, pipeline, rdp):
        if self.cache is not None:
            return
        if not self.op.children or len(self.op.children) == 0:
            _proc = self.get_function_spark(rdp)
            self.cache = _proc()
        else:
            _convert = None
            child_output = pipeline[self.op.children[0]].cache
            if isinstance(child_output, SparkDataFrame):
                if self.support_spark_dataframe:
                    _proc = self.get_function_spark(rdp)
                else:
                    _convert = SparkDataFrameToRDDConverter().get_function(rdp)
                    _proc = self.get_function_spark_rdd(rdp)
            elif isinstance(child_output, SparkRDD):
                if self.support_spark_rdd:
                    _proc = self.get_function_spark_rdd(rdp)
                else:
                    _convert = RDDToSparkDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_spark(rdp)
            elif isinstance(child_output, pd.DataFrame):
                if self.support_spark_rdd:
                    _convert = DataFrameToRDDConverter().get_function(rdp)
                    _proc = self.get_function_spark_rdd(rdp)
                else:
                    _convert = DataFrameToSparkDataFrameConverter().get_function(rdp)
                    _proc = self.get_function_spark(rdp)
            else:
                raise ValueError(f"child cache is not recognized {child_output}")
        
            if _convert:
                child_output = _convert(child_output)
            self.cache = _proc(child_output)
            #print(self.cache.take(1))

    def get_function_spark_rdd(self, rdp):
        actual_func = self.get_function_pd()
        def transform(iter, *args):
            for x in iter:
                yield actual_func(x, *args)
        def base_spark_feature_generator(rdd):
            return rdd.mapPartitions(transform)
        return base_spark_feature_generator


from .featuretools_adaptor import FeaturetoolsOperation

class BertDecodeOperation(FeaturetoolsOperation):
    def __init__(self, op_base):        
        super().__init__(op_base)

class TextFeatureGenerator(FeaturetoolsOperation):
    def __init__(self, op_base):        
        super().__init__(op_base)


from pyrecdp.primitives.generators import *
from .BasePipeline import BasePipeline
import logging
from pyrecdp.core.dataframe import DataFrameAPI

logging.basicConfig(format='%(asctime)s %(levelname)s:%(message)s', level=logging.ERROR, datefmt='%I:%M:%S')
logger = logging.getLogger(__name__)

class FeatureWrangler(BasePipeline):
    def __init__(self, dataset, label, supplementary_datasets = None, *args, **kwargs):
        super().__init__(dataset, label)
        self.data_profiler = [cls() for cls in feature_infer_list]
        # If we provided multiple datasets in this workload
        self.generators.append([DataframeConvertFeatureGenerator()])
        self.generators.append([cls() for cls in pre_feature_generator_list])
        self.generators.append([cls() for cls in transformation_generator_list])
        self.generators.append([cls() for cls in post_feature_generator_list])
        self.generators.append([DataframeTransformFeatureGenerator()])
        self.generators.append([cls() for cls in index_generator_list])
        self.generators.append([cls() for cls in encode_generator_list])
        self.generators.append([cls(final = True) for cls in final_generator_list])

        self.fit_analyze()

    def fit_analyze(self, *args, **kwargs):
        X = DataFrameAPI().instiate(self.feature_data)
        sampled_data = X.may_sample()
        child = list(self.pipeline.keys())[-1]
        max_id = child
        for generator in self.data_profiler:
            self.pipeline, child, max_id = generator.fit_prepare(self.pipeline, [child], max_id, sampled_data)
        super().fit_analyze(*args, **kwargs)


import os, requests
from tqdm import tqdm
import shutil
from pathlib import Path
import boto3

libpath = str(Path(__file__).parent.resolve())

class base_api:
    def __init__(self):        
        self.cache_dir = f"{libpath}/dataset_cache"
        os.makedirs(self.cache_dir, exist_ok = True)
       
    def download_s3(self, bucket, filename):
        to_save = f"{self.cache_dir}/{filename}"
        # check if miniconda exsists
        if os.path.exists(to_save):
            return to_save
        
        s3r = boto3.resource('s3', aws_access_key_id='AKIAYAY77NQAV5HDP7ID',
            aws_secret_access_key='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')
        buck = s3r.Bucket(bucket)
        buck.download_file(filename, to_save)
        return to_save
        
        
    def download_url(self, name, url):
        to_download = url
        to_save = f"{self.cache_dir}/{name}"
        # check if miniconda exsists
        if os.path.exists(to_save):
            return to_save

        with requests.get(to_download, stream=True) as r:
            # check header to get content length, in bytes
            total_length = int(r.headers.get("Content-Length"))
            
            # implement progress bar via tqdm
            with tqdm.wrapattr(r.raw, "read", total=total_length, desc="")as raw: 
                # save the output to a file
                with open(f"{to_save}", 'wb')as output:
                    shutil.copyfileobj(raw, output)
        return to_save


from .base import BaseFeatureGenerator as super_class
from pyrecdp.core import SeriesSchema
from pyrecdp.primitives.operations import Operation
import pandas as pd
import numpy as np
from pandas.api import types as pdt
import copy

def try_category(s):
    if pdt.is_categorical_dtype(s) and not pdt.is_bool_dtype(s):
        return False
    n_unique = s.nunique()
    total_len = len(s)
    threshold = (total_len / 5) if (total_len / 5) < 10000 else 10000
    if 1 <= n_unique <= threshold:
        return True
    return False

def try_datetime(s):
    if pdt.is_datetime64_any_dtype(s):
        return False
    if not pdt.is_string_dtype(s):
        return False
    if s.isnull().all():
        return False
    try:
        if len(s) > 500:
            # Sample to speed-up type inference
            result = s.sample(n=500, random_state=0)
        result = pd.to_datetime(result, errors='coerce')
        if result.isnull().mean() > 0.8:  # If over 80% of the rows are NaN
            return False
        else:
            return True
    except:
        return False
        
class TypeInferFeatureGenerator(super_class):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx, df):
        ret_pa_fields = []
        pa_schema = pipeline[children[0]].output
        for pa_field, feature_name in zip(pa_schema, df.columns):
            config = {}
            if try_category(df[feature_name]):
                config['is_categorical'] = True
            if try_datetime(df[feature_name]):
                config['is_datetime'] = True
            pa_field.copy_config_from(config)
            ret_pa_fields.append(pa_field)
        
        # append to pipeline
        cur_idx = max_idx + 1
        config = [x.dump() for x in ret_pa_fields]
        pipeline[cur_idx] = Operation(cur_idx, children, ret_pa_fields, op = 'type_infer', config = config)
        return pipeline, cur_idx, cur_idx
    

 
class TypeCheckFeatureGenerator(super_class):
    def __init__(self, final = False, **kwargs):
        super().__init__(**kwargs)
   
    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        config = {}
        for idx in range(len(pa_schema)):
            pa_field = pa_schema[idx]
            if pa_field.is_categorical_and_string:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.StringDtype())
                config[pa_field.name] = 'str'
            elif pa_field.is_categorical:
                pa_schema[idx] = SeriesSchema(pa_field.name, pd.Int32Dtype())
                config[pa_field.name] = 'int'
        
        cur_idx = max_idx
        #pipeline[cur_idx] = Operation(cur_idx, children, pa_schema, op = 'type_check', config = config)
        return pipeline, cur_idx, cur_idx



from .featuretools_adaptor import FeaturetoolsBasedFeatureGenerator
from featuretools.primitives import (
    Day,
    Month,
    Weekday,
    Year,
    Hour,
    PartOfDay
)

class DatetimeFeatureGenerator(FeaturetoolsBasedFeatureGenerator):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.op_list = [
            Day,
            Month,
            Weekday,
            Year,
            Hour,
            #PartOfDay()
        ]
        self.op_name = 'datetime_feature'            

    def fit_prepare(self, pipeline, children, max_idx):
        pa_schema = pipeline[children[0]].output
        for pa_field in pa_schema:
            if pa_field.is_datetime:
                self.feature_in.append(pa_field.name)
        return super().fit_prepare(pipeline, children, max_idx)
        